import	os
import	unittest
from	pathlib								import Path
from	shutil								import rmtree
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.hagrid				import EasySet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.planting.weeds		import TrimProbe
from	pygwarts.hagrid.planting.weeds		import SprigTrimmer
from	pygwarts.hagrid.bloom.weeds			import Efflorescence
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class EfflorescenceCase(EasySet):

	"""
		Flourishing and direct Efflorescence
		Single sprout
		Single bough
		Testing different variants
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.EASY_HANDLER_5): os.remove(cls.EASY_HANDLER_5)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.EASY_HANDLER_5)

		cls.fading1 = os.path.join(cls.EASY_SET_BOUGH, "fading1")
		cls.fading2 = os.path.join(cls.EASY_SET_BOUGH, "fading2")

		cls.notdst1 = os.path.join(cls.EASY_SET_BOUGH, "not.exist")
		cls.notdst2 = os.path.join(cls.EASY_SET_BOUGH, "not.exist1")
		cls.notdst3 = os.path.join(cls.EASY_SET_BOUGH, "not.exist2")
		cls.notdst4 = os.path.join(cls.EASY_SET_BOUGH, "not.exist3")
		cls.notsrc1 = os.path.join(cls.EASY_SET_SPROUT, "not.exist")
		cls.notsrc2 = os.path.join(cls.EASY_SET_SPROUT, "not.exist1")
		cls.notsrc3 = os.path.join(cls.EASY_SET_SPROUT, "not.exist2")
		cls.notsrc4 = os.path.join(cls.EASY_SET_SPROUT, "not.exist3")
		cls.notsrc5 = os.path.join(cls.EASY_SET_SPROUT, "non-existant folder")
		cls.notsrc6 = os.path.join(cls.notsrc5, "not.exist1")
		cls.notsrc7 = os.path.join(cls.notsrc5, "not.exist2")
		cls.notsrc8 = os.path.join(cls.notsrc5, "not.exist3")

		class Sakura(Tree):

			bough = cls.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				handler		= cls.EASY_HANDLER_5
				init_level	= 10

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	pass

		cls.sakura = Sakura


	def setUp(self):

		if	not os.path.isfile(self.fading1) : self.fmake(self.fading1, "this file must fade away")
		os.makedirs(self.fading2, exist_ok=True)




	def test_nobranches(self):

		self.sakura.loggy.init_name = "effloresce_nobranches"
		with self.assertLogs("effloresce_nobranches", 10) as case_loggy:

			self.test_case = self.sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.clean.blooming),0)
		self.assertNotIn("self.test_case", self.test_case.clean.blooming)
		self.assertNotIn(str(self.test_case), self.test_case.clean.blooming)
		self.assertIn(

			"DEBUG:effloresce_nobranches:Branches not found or sprout is not a string",
			case_loggy.output
		)








	def test_nosprout(self):

		self.sakura.loggy.init_name = "effloresce_nosprout"
		with self.assertLogs("effloresce_nosprout", 10) as case_loggy:

			self.test_case = self.sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case[0](

				self.test_case,
				self,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.clean.blooming),0)
		self.assertNotIn("self.test_case", self.test_case.clean.blooming)
		self.assertNotIn(str(self), self.test_case.clean.blooming)
		self.assertIn(

			"DEBUG:effloresce_nosprout:Branches not found or sprout is not a string",
			case_loggy.output
		)








	def test_nobough_1(self):

		self.sakura.loggy.init_name = "effloresce_nobough_1"
		with self.assertLogs("effloresce_nobough_1", 10) as case_loggy:

			self.test_case = self.sakura()
			self.test_case.branches = dict()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.clean.blooming),1)
		self.assertIn("self.test_case", self.test_case.clean.blooming)
		self.assertNotIn("trim", self.test_case.clean.blooming["self.test_case"])
		self.assertFalse(self.test_case.clean.blooming["self.test_case"]["bough"])
		self.assertIn(

			"ERROR:effloresce_nobough_1:Invalid tree self.test_case or no bough to effloresce",
			case_loggy.output
		)








	def test_nobough_2(self):

		self.sakura.loggy.init_name = "effloresce_nobough_2"
		with self.assertLogs("effloresce_nobough_2", 10) as case_loggy:

			self.test_case = self.sakura()
			self.test_case.branches = dict()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)
			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)
			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

			self.assertFalse(self.test_case.clean.blooming["self.test_case"]["bough"])
			self.test_case.clean.blooming["self.test_case"]["bough"] = True
			self.assertTrue(self.test_case.clean.blooming["self.test_case"]["bough"])

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.clean.blooming),1)
		self.assertIn("self.test_case", self.test_case.clean.blooming)
		self.assertNotIn("trim", self.test_case.clean.blooming["self.test_case"])
		self.assertFalse(self.test_case.clean.blooming["self.test_case"]["bough"])
		self.assertIn(

			"ERROR:effloresce_nobough_2:Invalid tree self.test_case or no bough to effloresce",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:effloresce_nobough_2:Invalid tree self.test_case or no bough to effloresce",
			case_loggy.output
		)
		self.assertEqual(

			case_loggy.output.count(
				"ERROR:effloresce_nobough_2:Invalid tree self.test_case or no bough to effloresce"
			), 2
		)
		self.assertEqual(

			case_loggy.output.count(
				"DEBUG:effloresce_nobough_2:Invalid tree self.test_case or no bough to effloresce"
			), 2
		)








	def test_notrim_1(self):

		self.sakura.loggy.init_name = "effloresce_notrim_1"
		with self.assertLogs("effloresce_notrim_1", 10) as case_loggy:

			self.test_case = self.sakura()
			self.test_case.branches = dict()
			self.test_case.trim = 42

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.clean.blooming),1)
		self.assertIn(str(self.test_case), self.test_case.clean.blooming)
		self.assertNotIn("bough", self.test_case.clean.blooming[str(self.test_case)])
		self.assertFalse(self.test_case.clean.blooming[str(self.test_case)]["trim"])
		self.assertIn(

			f"ERROR:effloresce_notrim_1:{self.test_case} doesn't implement trim",
			case_loggy.output
		)








	def test_notrim_2(self):

		self.sakura.loggy.init_name = "effloresce_notrim_2"
		with self.assertLogs("effloresce_notrim_2", 10) as case_loggy:

			self.test_case = self.sakura()
			self.test_case.branches = dict()
			self.test_case.trim = 42

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)
			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)
			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

			self.assertFalse(self.test_case.clean.blooming[str(self.test_case)]["trim"])
			self.test_case.clean.blooming[str(self.test_case)]["trim"] = True
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["trim"])

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.clean.blooming),1)
		self.assertIn(str(self.test_case), self.test_case.clean.blooming)
		self.assertNotIn("bough", self.test_case.clean.blooming[str(self.test_case)])
		self.assertFalse(self.test_case.clean.blooming[str(self.test_case)]["trim"])
		self.assertIn(

			f"ERROR:effloresce_notrim_2:{self.test_case} doesn't implement trim",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:effloresce_notrim_2:{self.test_case} doesn't implement trim",
			case_loggy.output
		)
		self.assertEqual(

			case_loggy.output.count(
				f"ERROR:effloresce_notrim_2:{self.test_case} doesn't implement trim"
			),2
		)
		self.assertEqual(

			case_loggy.output.count(
				f"DEBUG:effloresce_notrim_2:{self.test_case} doesn't implement trim"
			),2
		)








	def test_nocontrollers_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "nocontrollers_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("nocontrollers_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:nocontrollers_flourish:Controller for twigs not found",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nocontrollers_flourish:Controller for leafs not found",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nocontrollers_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nocontrollers_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nocontrollers_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nocontrollers_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_branches_1_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "branches_1_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = dict()

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("branches_1_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:branches_1_flourish:Bough \"{self.EASY_SET_BOUGH}\" not included for Efflorescence",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_1_flourish:Sprout \"{self.EASY_SET_SPROUT}\" not included for Efflorescence",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_1_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_1_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_1_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_1_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_branches_2_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "branches_2_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: tuple() }

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("branches_2_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:branches_2_flourish:Sprout \"{self.EASY_SET_SPROUT}\" not included for Efflorescence",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_2_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_2_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_2_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_2_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_branches_3_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "branches_3_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: list() }

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("branches_3_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:branches_3_flourish:Invalid sprouts mapping for \"{self.EASY_SET_BOUGH}\"",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_3_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_3_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_3_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:branches_3_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_branches_4_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "branches_4_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_SPROUT }

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("branches_4_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.clean.includable(self.EASY_SET_BOUGH)

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:branches_4_flourish:No branches included for Efflorescence",
			case_loggy.output
		)








	def test_wrong_include_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "wrong_include_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			# Looks like controller do have include field, but it's wrong types
			class leafs(SiftingController):

				include = (

					( rf"{self.EASY_SET_BOUGH}/.+" )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+" )
				)
			class twigs(SiftingController):

				include = (

					( rf"{self.EASY_SET_BOUGH}/.+" )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+" )
				)
			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}



			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("wrong_include_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.assertIn(

			"WARNING:wrong_include_flourish:"
			f"{self.test_case.twigs} include field improper, must be not empty list or tuple",
			case_loggy.output
		)
		self.assertIn(

			"WARNING:wrong_include_flourish:"
			f"{self.test_case.leafs} include field improper, must be not empty list or tuple",
			case_loggy.output
		)
		self.assertIn(
			"DEBUG:wrong_include_flourish:Include field for twigs not found", case_loggy.output
		)
		self.assertIn(
			"DEBUG:wrong_include_flourish:Include field for leafs not found", case_loggy.output
		)
		self.assertIn(
			f"DEBUG:wrong_include_flourish:Bough \"{self.dst_pros_folder}\" not thrived", case_loggy.output
		)
		self.assertIn(
			f"DEBUG:wrong_include_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived", case_loggy.output
		)
		self.assertIn(
			f"DEBUG:wrong_include_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived", case_loggy.output
		)
		self.assertIn(
			f"DEBUG:wrong_include_flourish:Bough \"{self.dst_cons_folder}\" not thrived", case_loggy.output
		)








	def test_probe_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "probe_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			# This is the example of a rule exception. The sifting basically designed to decalre
			# "twigs" and "leafs" separately under dispatcher (Flourish) object (main sifting)
			# and for every tree/copse somewhere else as many as need. But this situation is classic
			# cause it is single Tree that need effloresce, so for this to be done, the Tree object
			# must have controllers by itself, it wont be able to reach controllers under the Flourish.
			# So they are declared under the Tree. If Floursih will happen to have itself controllers,
			# main sifting will proceed with it. But in this situation, when there is no need in
			# main sifting, it will happen anyway, cause Flourish has access to the Tree's controllers.
			# So actually, includables for Efflorescence will not affect main sifting, just will take
			# some time, but the idea is the ability to put main siftables to the Tree.
			class leafs(SiftingController):
				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)
			class twigs(SiftingController):
				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)
			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("probe_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),1)
			self.assertIn(str(self.test_case), self.test_case.clean.blooming)
			self.assertIn("trim", self.test_case.clean.blooming[str(self.test_case)])
			self.assertIn("bough", self.test_case.clean.blooming[str(self.test_case)])
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["trim"])
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["bough"])


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertTrue(os.path.isdir(self.EASY_SET_BOUGH))
		self.assertTrue(os.path.isfile(self.fading1))
		self.assertTrue(os.path.isdir(self.fading2))


		self.assertIn(f"INFO:probe_flourish:Leaf \"{self.fading1}\" trim probe", case_loggy.output)
		self.assertIn(f"INFO:probe_flourish:Twig \"{self.fading2}\" trim probe", case_loggy.output)
		self.assertNotIn(
			f"DEBUG:probe_flourish:Bough \"{self.dst_pros_folder}\" not thrived", case_loggy.output
		)
		self.assertNotIn(
			f"DEBUG:probe_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived", case_loggy.output
		)
		self.assertNotIn(
			f"DEBUG:probe_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived", case_loggy.output
		)
		self.assertNotIn(
			f"DEBUG:probe_flourish:Bough \"{self.dst_cons_folder}\" not thrived", case_loggy.output
		)








	def test_probe_no_thrive_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "probe_not_thrive_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class leafs(SiftingController):
				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)
			class twigs(SiftingController):
				include = (

					( rf"({self.EASY_SET_BOUGH}|{self.EASY_SET_SPROUT})/.+", )
					if os.name == "posix" else
					(
						r"(%s|%s)\\.+"%(

							self.EASY_SET_BOUGH.replace("\\", "\\\\"),
							self.EASY_SET_SPROUT.replace("\\", "\\\\")
						),
					)
				)

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("probe_not_thrive_flourish", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),1)
			self.assertIn(str(self.test_case), self.test_case.clean.blooming)
			self.assertIn("trim", self.test_case.clean.blooming[str(self.test_case)])
			self.assertIn("bough", self.test_case.clean.blooming[str(self.test_case)])
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["trim"])
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["bough"])


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertTrue(os.path.isdir(self.EASY_SET_BOUGH))
		self.assertTrue(os.path.isfile(self.fading1))
		self.assertTrue(os.path.isdir(self.fading2))


		self.assertIn(f"INFO:probe_not_thrive_flourish:Leaf \"{self.fading1}\" trim probe", case_loggy.output)
		self.assertIn(f"INFO:probe_not_thrive_flourish:Twig \"{self.fading2}\" trim probe", case_loggy.output)
		self.assertIn(

			f"DEBUG:probe_not_thrive_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:probe_not_thrive_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:probe_not_thrive_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:probe_not_thrive_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_bad_probe(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "bad_probe_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class leafs(SiftingController):

				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)
			class twigs(SiftingController):


				include = (

					( rf"({self.EASY_SET_BOUGH}|{self.EASY_SET_SPROUT})/.+", )
					if os.name == "posix" else
					(
						r"(%s|%s)\\.+"%(

							self.EASY_SET_BOUGH.replace("\\", "\\\\"),
							self.EASY_SET_SPROUT.replace("\\", "\\\\")
						),
					)
				)

			class trim(TrimProbe):		pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("bad_probe_flourish", 10) as case_loggy:

			self.test_case = Sakura()

			self.assertTrue(len(self.test_case) == 1)
			self.assertTrue(str(self.test_case[0]) == "Sakura.clean")
			self.assertTrue(isinstance(self.test_case[0], Efflorescence))
			self.assertTrue(hasattr(self.test_case.clean, "blooming"))
			self.assertTrue(isinstance(self.test_case.clean.blooming, dict))
			self.assertEqual(len(self.test_case.clean.blooming),0)

			self.test_case.trim(

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				Path(self.EASY_SET_SPROUT).joinpath("non-existiest file"),
				Path(self.EASY_SET_BOUGH)
			)


		self.assertEqual(len(self.test_case.clean.blooming),0)
		self.assertNotIn(str(self.test_case), self.test_case.clean.blooming)
		self.assertIn(

			"WARNING:bad_probe_flourish:Bad sprig "
			f"\"{os.path.join(self.EASY_SET_SPROUT, 'non-existiest file')}\""
			" to probe",
			case_loggy.output
		)








	def test_trim_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "trim_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class leafs(SiftingController):

				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)

			class twigs(SiftingController):

				include = (

					( rf"({self.EASY_SET_BOUGH}|{self.EASY_SET_SPROUT})/.+", )
					if os.name == "posix" else
					(
						r"(%s|%s)\\.+"%(

							self.EASY_SET_BOUGH.replace("\\", "\\\\"),
							self.EASY_SET_SPROUT.replace("\\", "\\\\")
						),
					)
				)

			class trim(SprigTrimmer):	pass
			class clean(Efflorescence):	branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		self.assertTrue(os.path.isfile(self.fading1))
		self.assertTrue(os.path.isdir(self.fading2))


		with self.assertLogs("trim_flourish", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),1)
			self.assertIn(str(self.test_case), self.test_case.clean.blooming)
			self.assertIn("trim", self.test_case.clean.blooming[str(self.test_case)])
			self.assertIn("bough", self.test_case.clean.blooming[str(self.test_case)])
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["trim"])
			self.assertTrue(self.test_case.clean.blooming[str(self.test_case)]["bough"])


		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertTrue(os.path.isdir(self.EASY_SET_BOUGH))
		self.assertFalse(os.path.isfile(self.fading1))
		self.assertFalse(os.path.isdir(self.fading2))


		self.assertIn(f"INFO:trim_flourish:Trimmed leaf \"{self.fading1}\"", case_loggy.output)
		self.assertIn(f"INFO:trim_flourish:Trimmed twig \"{self.fading2}\"", case_loggy.output)
		self.assertIn(

			f"DEBUG:trim_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:trim_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:trim_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:trim_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_raise_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "raise_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class leafs(SiftingController):

				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)
			class twigs(SiftingController):

				include = (

					( rf"({self.EASY_SET_BOUGH}|{self.EASY_SET_SPROUT})/.+", )
					if os.name == "posix" else
					(
						r"(%s|%s)\\.+"%(

							self.EASY_SET_BOUGH.replace("\\", "\\\\"),
							self.EASY_SET_SPROUT.replace("\\", "\\\\")
						),
					)
				)

			class trim(Transmutable):
				def __call__(self, origin :Tree, sprout :str, branch :Path, sprig :str, bough :Path):

					target_branch = bough.joinpath(sprig)
					raise	PermissionError(f"Failed with \"{target_branch}\"")

			class clean(Efflorescence): branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("raise_flourish", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()

			self.assertEqual(len(self.test_case.clean.blooming),0)


		self.assertTrue(os.path.isdir(self.EASY_SET_BOUGH))
		self.assertTrue(os.path.isfile(self.fading1))
		self.assertTrue(os.path.isdir(self.fading2))


		self.assertNotIn(f"INFO:raise_flourish:Trimmed leaf \"{self.fading1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:raise_flourish:Trimmed twig \"{self.fading2}\"", case_loggy.output)
		self.assertIn(

			"WARNING:raise_flourish:Failed to effloresce "
			f"\"{self.fading1}\" due to PermissionError: "
			f"Failed with \"{self.fading1}\"",
			case_loggy.output
		)
		self.assertIn(

			"WARNING:raise_flourish:Failed to effloresce "
			f"\"{self.fading2}\" due to PermissionError: "
			f"Failed with \"{self.fading2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:raise_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:raise_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:raise_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:raise_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








	def test_nosprigs_flourish(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "nosprigs_flourish"
				handler		= self.EASY_HANDLER_5
				init_level	= 10

			class leafs(SiftingController):

				include = (

					( rf"{self.EASY_SET_BOUGH}/.+", )
					if os.name == "posix" else
					( self.EASY_SET_BOUGH.replace("\\", "\\\\") + r"\\.+", )
				)

			class twigs(SiftingController):

				include = (

					( rf"({self.EASY_SET_BOUGH}|{self.EASY_SET_SPROUT})/.+", )
					if os.name == "posix" else
					(
						r"(%s|%s)\\.+"%(

							self.EASY_SET_BOUGH.replace("\\", "\\\\"),
							self.EASY_SET_SPROUT.replace("\\", "\\\\")
						),
					)
				)

			class trim(SprigTrimmer):	pass
			class clean(Efflorescence): branches = { self.EASY_SET_BOUGH: ( self.EASY_SET_SPROUT, )}

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		if	os.path.isfile(self.fading1): os.remove(self.fading1)
		if	os.path.isdir(self.fading2)	: rmtree(self.fading2)

		with self.assertLogs("nosprigs_flourish", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()

		self.assertEqual(len(self.test_case.clean.blooming),0)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertTrue(os.path.isdir(self.EASY_SET_BOUGH))
		self.assertFalse(os.path.isfile(self.fading1))
		self.assertFalse(os.path.isdir(self.fading2))


		self.assertNotIn(f"INFO:nosprigs_flourish:Trimmed leaf \"{self.fading1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:nosprigs_flourish:Trimmed twig \"{self.fading2}\"", case_loggy.output)
		self.assertIn(
			f"DEBUG:nosprigs_flourish:No sprigs to effloresce \"{self.EASY_SET_BOUGH}\"", case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nosprigs_flourish:Bough \"{self.dst_pros_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nosprigs_flourish:Bough \"{self.dst_redundant_1_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nosprigs_flourish:Bough \"{self.dst_redundant_2_folder}\" not thrived",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:nosprigs_flourish:Bough \"{self.dst_cons_folder}\" not thrived",
			case_loggy.output
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing							import Optional
from typing							import Callable
from pygwarts.magical.spells		import flagrate
from pygwarts.irma.access.inducers	import AccessInducer
from pygwarts.irma.access.volume	import LibraryVolume








class RegisterRecapInducer(AccessInducer):

	"""
		Implementation of inducer, that accesses provided "volume" argument LibraryVolume object with
		AccessHandler object it is assigned to, so it must be it's "_UPPER_LAYER", and returns corresponding
		mapped "recap" value, which might be any type and represents the object AccessHandler maintains for
		"volume". Processes obtained value with special field "filter", which must be callable that accepts
		two arguments - current AccessInducer object and the value it processing. It is must be noted, that
		"filter" invocation implemented according to "filter" assigned during declaration of AccessInducer
		class, so it is instantiated. If "filter" will be assigned to already instantiated AccessInducer
		object, it must account for accepting the value argument only. It doesn't restricted for "filter"
		what value to return, cause it anyway will be converted to string. Returns either obtained, and
		perhaps processed, value if it is originally not None or "filter" returned not None. Otherwise
		returns None. There is no Exception handling for "filter" invocation.
	"""

	filter :Optional[Callable[[AccessInducer,object],str | None]]

	def __call__(self, volume :LibraryVolume) -> str | None :
		if	(recap := self.get_register_recap(volume)) is not None:

			if		callable(getattr(self, "filter", None)): recap = self.filter(recap)
			else:	self.loggy.debug("Filter not applied")
			if		recap is not None : return str(recap)








class RegisterRecapAccumulatorInducer(AccessInducer):

	"""
		Implementation of inducer, that accesses provided "volume" argument LibraryVolume object with
		AccessHandler object it is assigned to, so it must be it's "_UPPER_LAYER", and returns corresponding
		mapped "recap" value, which might be a list that represents the data AccessHandler maintains for
		"volume". Processes current "recap" list to obtain another list of unique items, if "unique" boolean
		field is set to True. Converts all items in "recap" to strings and join them with "joint" field
		value, that converted to string too. Returns resulted joined string. Returns None if either "recap"
		is not a list or empty list.
	"""


	joint	:Optional[str]
	unique	:Optional[bool]


	def __call__(self, volume :LibraryVolume) -> str | None :


		if	isinstance(recap := self.get_register_recap(volume), list):
			if	getattr(self, "unique", False):


				uniqs	= set()
				accum	= [ acc for acc in map(str,recap) if not (acc in uniqs or uniqs.add(acc)) ]
			else:
				accum	= list(map(str,recap))


			induce	= str(getattr(self, "joint", " ")).join(accum)
			count	= len(induce)


			self.loggy.debug(f"Accumulated {count} symbol{flagrate(count)} string")
			return induce or None










import	re
from	pathlib								import Path
from	typing								import List
from	typing								import Tuple
from	typing								import Literal
from	typing								import Optional
from	collections.abc						import Sequence
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.spells				import patronus








class SiftingController(Transmutable):

	"""
		hagrid.cultivation core class that implements filtering of folders/files to be "planted"
		(processed).
		Consists of "controller" (the class itself) which serves as a dispatcher and the Sieve class,
		which does the actual "sifting" (filtering).
		The "controller" part operates on special fields to be provided - "include" and "exclude", which
		refers to corresponding "folders/files to be included" and "folders/files to be excluded". This
		fields might be populated by either compiled Patterns or compilable strings, which both must
		represent the desired items to be excluded/included for "planting". The item's precision is a
		tricky thing, like it is allowed to use any wild cards, even just ".+", but the end effect
		of such filtering might depend on the other factors of "planting" organization. In initiation time
		this fields will be searched for and handled by special method "pmake_field" to provide another
		corresponding fields "include_field" and "exclude_field", which will be used as arguments in
		Sieve part. It is possible to attach directly a list of strings/Patterns or new
		include_field/exclude_field processed by call to "pmake_field" at any time, or just pass some as
		optional key-word-only arguments to __call__ (the last one option always get priority over fields).
		The "controller" invocation must be granted "sprigs" argument - list of folders/files as Path
		objects to be filtered, as first and only positional argument. Additional optional key-word-only
		arguments might be provided:
			popping	-	boolean flag, which if True, allows to alter "sprigs" list by removing
						folders/files that was sifted out (this is the main reason why "sprigs" must always
						be the list, cause the original design implies "sprigs" to be obtained from object,
						which will use this altered list, so this modifications must are able to affect such
						object behavior; in other words, when popping used in *"global" folders sifting by
						default, every folder that was popped from "sprigs" list will not be processed
						further, hence that folder content will not be considered).
						False by default;
			thriving -	boolean flag, which if True, allows additional matching for "include_field". This
						matching suggests convenient usage of "include_field", such providing regular
						expression with wildcards at the end, but a real path in the beginning. When some
						parent folder of desired "includable" is considered, simple matching will not work,
						cause Sieve relies on "fullmatch" only, and as result that parent folder will be
						sifted out and desired child item won't be reached. For such cases "thriving" flag
						is toggled on by default in *"global" folders sifting for "include_field" processing
						in Sieve, so special method "thriving_check" will check if current branch will
						return True for any "includable" pattern string "startswith" method, to make final
						decision about sifting.
						False by default.
		The "controller" __call__ will return a new list of filtered folders/files as Path objects in case
		of success, or the original "sprigs" list of Path objects otherwise.
		The original design of "planting" require "SiftingController" for folders to be declared as "twigs"
		and for files as "leafs". Only this names will be searched by default in hagrid to handle any sifting.
		The very important thing, that hagrid has 2 basic "siftings": *"global" and "local".
		First one implies filtering for every Tree to be "planted", while "local" is an auxiliary option.
		In their difference lies the necessity of isolated declaration - "global" must have it's very own
		"controllers", which are reachable only for *object that implements "global" sifting; "local" might
		be declared anywhere except *"global" object, with only restriction that it must be reachable for
		the Tree objects it is belong to. As the default behavior of *"global" object and Tree objects
		include the phase of automatic sifting invocation, Tree objects might be encapsulated to Copse object
		to provide common "controllers", and also to isolate some Tree objects form another and from *"global"
		object, that Copse might be encapsulated in any more Copse objects. Last idea will allow isolate
		"controllers" of Tree objects from *"global" object when the last one doesn't have it's own, and
		they will be search by escalation. Anyway, the situation when *"global" object and Tree object share
		"controllers" is not dangerous, and, when fields are well organized, might only lead to redundant
		sifting for already sifted items. "controllers" for both "global" and "local" might be omitted.

		*	- refer to hagrid.planting.Flourish
	"""

	include	:Optional[Tuple[str | re.Pattern]]
	exclude	:Optional[Tuple[str | re.Pattern]]


	def pmake_field(self, entities :Sequence[str | re.Pattern]) -> List[re.Pattern] :

		"""
			SiftingController helper method, that accepts sequence of entities to produce list of Patterns
			objects to be used for SiftingController.Sieve. Entities must be whether compiled Patterns, or
			compilable strings. Always return list, which might be empty in case of empty "entities" or
			if no entity satisfied the above condition, with every Exception that raised being logged.
		"""

		self.loggy.debug(f"Making field from {len(entities)} entities")
		patterns = list()


		for candidate in entities:
			match candidate:


				case re.Pattern():

					patterns.append(candidate)
					self.loggy.debug(f"Added pattern \"{candidate.pattern}\"")


				case str():
					try:

						patterns.append(re.compile(candidate))
						self.loggy.debug(f"Made pattern \"{candidate}\"")


					except	Exception as E:

						self.loggy.warning(f"{self} didn't made pattern \"{candidate}\" due to {patronus(E)}")
				case _:	self.loggy.warning(f"Unpatternable type {type(candidate)}")
		return	patterns




	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


		if	hasattr(self, "exclude"):
			if	isinstance(self.exclude, list | tuple) and len(self.exclude):
				if	(excludables := self.pmake_field(self.exclude)):

					self.exclude_field = list(excludables)
					self.loggy.debug(f"Exclude field length: {len(self.exclude_field)}")
			else:	self.loggy.warning(f"{self} exclude field improper, must be not empty list or tuple")
		else:		self.loggy.debug("Exclude field not implemented for sifting")




		if	hasattr(self, "include"):
			if	isinstance(self.include, list | tuple) and len(self.include):
				if	(includables := self.pmake_field(self.include)):

					self.include_field = list(includables)
					self.loggy.debug(f"Include field length: {len(self.include_field)}")
			else:	self.loggy.warning(f"{self} include field improper, must be not empty list or tuple")
		else:		self.loggy.debug("Include field not implemented for sifting")








	def __call__(
					self,
					sprigs		:List[Path],
					*,
					excludables	:Sequence[re.Pattern]	=None,
					includables	:Sequence[re.Pattern]	=None,
					popping		:bool					=False,
					thriving	:bool					=False
				)-> List[Path] :


		if	callable((sifted := getattr(self, "Sieve", None))):
			if	isinstance(sprigs, list) and not (sifted_sprigs := list()):


				for sprig in sprigs:
					if	isinstance(sprig, Path):

						siftable = str(sprig)
						self.loggy.debug(f"Considering siftable \"{siftable}\"")


						# The optional key-word-only arguments always get priority over corresponding
						# fields that might be omitted or created improperly, so the ability of using
						# different excludables/includables grants flexibility. It is important to note,
						# that excluding always goes before including, even in arguments.
						if	not sifted(
							siftable,

							exclude=excludables if excludables is not None else
							getattr(self, "exclude_field", None),

							include=includables if includables is not None else
							getattr(self, "include_field", None),

							thriving=thriving
						):
							sifted_sprigs.append(sprig)
							self.loggy.debug(f"Passed sprig \"{sprig.name}\"")
						else:
							self.loggy.debug(f"Sifted sprig \"{sprig.name}\"")
					else:	self.loggy.warning(f"{self} found invalid sprig \"{sprig}\"")




				# Removing instead of actually popping the sifted out sprigs from original list.
				# It is important, that current logic relies on maintaining "sprigs" list, that
				# was obtained from some object, which will use modified list, and it is incorrect
				# to just reassign "sprigs" to sifted sprigs.
				if	popping:
					for thorn in [ sprig for sprig in sprigs if sprig not in sifted_sprigs ]:

						sprigs.remove(thorn)
				return	sifted_sprigs
			else:		self.loggy.warning(f"{self} found invalid sprigs type {type(sprigs)}")
		else:			self.loggy.debug("Sieve not implemented for current controller")


		# Sifting gone wrong, so returning origingal sprigs
		return	sprigs








	class Sieve(Transmutable):

		"""
			hagrid.cultivation.SiftingController implementation of actual "sifting" (folders/files filtering).
			This object relies on 2 lists "exclude" and "include", which must content Patterns to matched
			with provided argument "siftable", which must be a string that represents folder/file absolute
			path to be processed. Those 2 lists are mandatory key-word-only arguments, will be iterated over
			to "fullmatch" siftable with any "excludable/includable", and "excludables" has a priority to be
			processed first. If "excludables" and "includables" are not the subjects to be provided, it must
			be provided as None both, as "controller" does by default. Last mandatory key-word-only argument
			is boolean flag "thriving", which if True, allows additional matching for "includables" by use of
			special method "thriving_check", that will check if current siftable will return True for any
			"includable" pattern string "startswith" method, to make final decision about "siftable". This
			matching suggests convenient usage of "includables", such providing regular expression with
			wildcards at the end, but a real path in the beginning. When some parent folder of desired
			"includable" is considered, simple matching will not work, cause Sieve relies on "fullmatch"
			only, and as result that parent folder will be sifted out and desired child item won't be reached.
			For such cases "thriving" option might be used, and it is used by default in *"global" folders
			sifting.
			This object's __call__ return value is boolean, which means whether provided "siftable"
			argument if sifted (filtered) out or not, so "True" is the signal for caller to NOT "plant" that
			siftable. This boolean might be obtained as follows:
				True	-	at the very beginning of __call__ if provided "siftable" is not a string.
				True	-	in "excludables" iterating, when provided "siftable" have a match.
				False	-	in not empty "includables" iterating, when provided "siftable" have a match.
				False	-	after "includables" processing, when "includables" is None.
				True	-	default return value, which will be returned by default if none of below
							happened, or if there was any trouble during "includables" iteration, as
							whether to include "siftable" is a picky moment. The "excludables" iterating
							has the opposite logic, that when any trouble occurs, "excludables" skipped,
							just like when it's None or empty.

			* - refer to hagrid.planting.Flourish
		"""

		def thriving_check(self, includable :re.Pattern, siftable :str) -> Literal[True] | None :

			"""
				Sieve helper method to check if "siftable" is a substring of "includable".
				Suggests convenient usage of "include_field", such providing regular expression with
				wildcards at the end, but a real path in the beginning. When some parent folder of desired
				"includable" is considered, simple matching will not work, cause Sieve relies on "fullmatch"
				only, and as result that parent folder will be sifted out and desired child item won't be
				reached. When current method return True it is a signal for caller, that "siftable" is a
				path to "includable" so it must not be sifted (filtered) out.
			"""

			if	isinstance(includable, re.Pattern):
				if	isinstance(siftable, str):
					if	includable.pattern.startswith(siftable):

						self.loggy.debug("Thriving check passed")
						return	True




		def __call__(
						self,
						siftable	:str,
						*,
						exclude		:List[re.Pattern] | None,
						include		:List[re.Pattern] | None,
						thriving	:bool
					)-> bool		:


			if	not isinstance(siftable, str):

				self.loggy.debug(f"Got invalid siftable \"{siftable}\"")
				return	True




			if	exclude is not None:
				if	isinstance(exclude, list):


					for excludable in exclude:


						if	isinstance(excludable, re.Pattern):
							if	excludable.fullmatch(siftable):

								self.loggy.debug(f"Exclude field triggered, \"{siftable}\" sifted")
								return	True
						else:	self.loggy.debug(f"Got invalid excludable \"{excludable}\"")
					else:		self.loggy.debug("Exclude field passed")
				else:			self.loggy.debug(f"Got invalid exclude field")
			else:				self.loggy.debug("Exclude field skipped")




			if	include is not None:
				if	isinstance(include, list):


					for includable in include:


						if	isinstance(includable, re.Pattern):
							if	includable.fullmatch(siftable) or (
								thriving and self.thriving_check(includable, siftable)
								):

								self.loggy.debug(f"Include field triggered, \"{siftable}\" passed")
								return	False
						else:	self.loggy.debug(f"Got invalid includable \"{includable}\"")
					else:		self.loggy.debug("Include field not passed")
				else:			self.loggy.debug("Got invalid include field")
			else:


				self.loggy.debug("Include field skipped")
				return	False
			return		True










import	unittest
from	pygwarts.tests.hedwig.unit_hedwig_mail_utils	import *
from	pygwarts.tests.hedwig.unit_hedwig_mail_letter	import *
from	pygwarts.tests.hedwig.unit_hedwig_mail_builder	import *
from	pygwarts.tests.hedwig.unit_hedwig_mail_client	import *
from	pygwarts.tests.hedwig.unit_hedwig_mail_smtp		import *








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pygwarts.tests.hagrid				import MediumSet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.thrivables			import Copse
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.bloom.leafs			import Transfer
from	pygwarts.hagrid.planting.leafs		import LeafMove
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class MultiTransferingCase(MediumSet):

	"""
		GrowingPeel, Transfer, moving graft
		File moving handler with all boughs filtered to collect corresponding files
		Triple sprout copse
		Triple bough copse
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.MEDIUM_HANDLER_2): os.remove(cls.MEDIUM_HANDLER_2)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.MEDIUM_HANDLER_2)
	def setUp(self): super().setUpClass()
	def test_filtered_flourish(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_2
				init_name	= "filtered_flourish"
				init_level	= 10

			class First(Tree):

				bough	= self.MEDIUM_SET_BOUGH_1
				class leafs(SiftingController):
					include	= ( r".+/.+\.task$", ) if os.name == "posix" else ( r".+\\.+\.task$", )

			class Second(Tree):

				bough	= self.MEDIUM_SET_BOUGH_2
				class leafs(SiftingController):
					include	= (

						( r".+/(onion|milk|cheese|meat|chicken|pie|bread|crumb.+)$", )
						if os.name == "posix" else
						( r".+\\(onion|milk|cheese|meat|chicken|pie|bread|crumb.+)$", )
					)

			class Third(Tree):

				bough	= self.MEDIUM_SET_BOUGH_3
				class leafs(SiftingController):

					include	= (

						( rf"{self.MEDIUM_SET_SPROUT_1}/men due/.+", )
						if os.name == "posix" else
						( self.MEDIUM_SET_SPROUT_1.replace("\\", "\\\\") + r"\\men due\\.+", )
					)
					exclude	= (

						( rf"{self.MEDIUM_SET_SPROUT_1}/men due/girls cries/.+", )
						if os.name == "posix" else
						( self.MEDIUM_SET_SPROUT_1.replace("\\", "\\\\") + r"\\men due\\girls cries\\.+", )
					)

			@GrowingPeel
			class graft(LeafMove):	pass
			class files(Transfer):	pass

			@fssprout(self.MEDIUM_SET_SPROUT_3)	# Order does matter cause every sprout pack it'self
			@fssprout(self.MEDIUM_SET_SPROUT_2)	# before previous sprout, so the nearest to dispatcher
			@fssprout(self.MEDIUM_SET_SPROUT_1)	# sprout will be processed first.
			class sync(Flourish):	pass




		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_1))
		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_2))
		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_3))


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		with self.assertLogs("filtered_flourish", 10) as case_loggy:

			self.test_case = Forest()
			self.test_case.sync()




		self.no_loggy_levels(case_loggy.output, 30,40,50)




		# First sprout
		self.assertFalse(os.path.isfile(self.f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.f_mss1_md_paper))
		self.assertFalse(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertFalse(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertFalse(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertFalse(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertFalse(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertFalse(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		# First sprout first bough
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_md))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_onion))

		# Second sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_easyphilosophizing))

		# Third sprout first bough
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_f))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_o))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_t))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_t_bc))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_f_milk))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_f_cheese))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_f_meat))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_o_chicken))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_o_pie))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb69))




		# First sprout second bough
		self.assertTrue(os.path.isdir(self.tg2_d_mss1_md))
		self.assertTrue(os.path.isdir(self.tg2_d_mss1_md_gc))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_goodbad))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_paper))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_scissors))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg2_f_mss1_md_gc_onion))

		# Second sprout second bough
		self.assertFalse(os.path.isdir(self.tg2_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_easyphilosophizing))

		# Third sprout second bough
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_f))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_o))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_t))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_t_bc))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb69))




		# First sprout third bough
		self.assertTrue(os.path.isdir(self.tg3_d_mss1_md))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_md_gc))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_scissors))
		self.assertFalse(os.path.isfile(self.tg3_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg3_f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.tg3_f_mss1_md_gc_onion))

		# Second sprout third bough
		self.assertFalse(os.path.isdir(self.tg3_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_easyphilosophizing))

		# Third sprout third bough
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_f))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_o))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_t))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_t_bc))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_f_milk))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_f_cheese))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_f_meat))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_o_chicken))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_o_pie))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb69))
















	def test_unfiltered_flourish(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_2
				init_name	= "unfiltered_flourish"
				init_level	= 10

			class First(Tree):

				bough	= self.MEDIUM_SET_BOUGH_1
				class leafs(SiftingController):
					include	= ( r".+/.+\.task$", ) if os.name == "posix" else ( r".+\\.+\.task$", )

			class Second(Tree):

				bough	= self.MEDIUM_SET_BOUGH_2
				class leafs(SiftingController):
					include	= (

						( r".+/(onion|milk|cheese|meat|chicken|pie|bread|crumb.+)$",)
						if os.name == "posix" else
						( r".+\\(onion|milk|cheese|meat|chicken|pie|bread|crumb.+)$",)
					)

			class Third(Tree):

				bough	= self.MEDIUM_SET_BOUGH_3
				class leafs(SiftingController):	pass

			@GrowingPeel
			class graft(LeafMove):				pass
			class files(Transfer):				pass

			@fssprout(self.MEDIUM_SET_SPROUT_3)	# Order does matter cause every sprout pack it'self
			@fssprout(self.MEDIUM_SET_SPROUT_2)	# before previous sprout, so the nearest to dispatcher
			@fssprout(self.MEDIUM_SET_SPROUT_1)	# sprout will be processed first.
			class sync(Flourish):				pass




		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_1))
		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_2))
		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_3))


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		with self.assertLogs("unfiltered_flourish", 10) as case_loggy:

			self.test_case = Forest()
			self.test_case.sync()




		self.no_loggy_levels(case_loggy.output, 30,40,50)




		# First sprout
		self.assertFalse(os.path.isfile(self.f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.f_mss1_md_paper))
		self.assertFalse(os.path.isfile(self.f_mss1_md_scissors))
		self.assertFalse(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertFalse(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertFalse(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertFalse(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertFalse(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertFalse(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertFalse(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		# First sprout first bough
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_md))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_onion))

		# Second sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.tg1_f_mss2_c_easyphilosophizing))

		# Third sprout first bough
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_f))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_o))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_t))
		self.assertFalse(os.path.isdir(self.tg1_d_mss3_k_t_bc))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_f_milk))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_f_cheese))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_f_meat))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_o_chicken))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_o_pie))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb69))




		# First sprout second bough
		self.assertTrue(os.path.isdir(self.tg2_d_mss1_md))
		self.assertTrue(os.path.isdir(self.tg2_d_mss1_md_gc))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg2_d_mss1_c_goodbad))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_paper))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_scissors))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg2_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg2_f_mss1_md_gc_onion))

		# Second sprout second bough
		self.assertFalse(os.path.isdir(self.tg2_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg2_f_mss2_c_easyphilosophizing))

		# Third sprout second bough
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_f))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_o))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_t))
		self.assertTrue(os.path.isdir(self.tg2_d_mss3_k_t_bc))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.tg2_f_mss3_k_t_bc_crumb69))




		# First sprout third bough
		self.assertTrue(os.path.isdir(self.tg3_d_mss1_md))
		self.assertTrue(os.path.isdir(self.tg3_d_mss1_md_gc))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg3_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.tg3_f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.tg3_f_mss1_md_gc_onion))

		# Second sprout third bough
		self.assertFalse(os.path.isdir(self.tg3_d_mss2_c))
		self.assertTrue(os.path.isfile(self.tg3_f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.tg3_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg3_f_mss2_c_easyphilosophizing))

		# Third sprout third bough
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_f))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_o))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_t))
		self.assertFalse(os.path.isdir(self.tg3_d_mss3_k_t_bc))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_f_milk))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_f_cheese))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_f_meat))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_o_chicken))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_o_pie))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg3_f_mss3_k_t_bc_crumb69))








if	__name__ == "__main__" : unittest.main(verbosity=2)







